/* This file is part of smap.
   Copyright (C) 2006, 2007, 2010 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include "smap.h"

char *progname = "[progname is not set, please report]";
char *log_tag;
int log_facility = LOG_DAEMON;
int log_to_stderr = -1;
int smap_trace_option;

smap_ostream_t smap_error_str;
smap_ostream_t smap_debug_str;
smap_ostream_t smap_trace_str;

void
smap_verror(const char *fmt, va_list ap)
{
	if (smap_ostream_vprintf(smap_error_str, fmt, ap) < 0) {
		vfprintf(stderr, fmt, ap);
		fputc('\n', stderr);
		fprintf(stderr,
			"INTERNAL ERROR: smap_error_str not "
			"initialized, please report\n");
		abort();
	}
	smap_ostream_write(smap_error_str, "\n", 1);
	smap_ostream_flush(smap_error_str);
}

void
smap_error(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	smap_verror(fmt, ap);
	va_end(ap);
}


static void
flush_stderr(const char *pfx, void *data, char *buf)
{
	if (pfx)
		fprintf(stderr, "%s: %s: %s\n", progname, pfx, buf);
	else
		fprintf(stderr, "%s: %s\n", progname, buf);
}

static void
flush_stderr_err(void *data, char *buf)
{
	flush_stderr(NULL, data, buf);
}

static void
flush_stderr_dbg(void *data, char *buf)
{
	flush_stderr("DEBUG", data, buf);
}

static void
flush_stderr_trc(void *data, char *buf)
{
	flush_stderr("TRACE", data, buf);
}

static void
flush_syslog(void *data, char *buf)
{
	int prio = (int)data;
	syslog(prio, "%s", buf);
}

void
set_program_name(char *name)
{
	char *p = strrchr(name, '/');
	if (p)
		p++;
	else
		p = name;
	if (strncmp(p, "lt-", 3) == 0)
		p += 3;
	progname = p;
}

void
smap_log_init()
{
	smap_ostream_destroy(&smap_error_str);
	smap_ostream_destroy(&smap_debug_str);
	smap_ostream_destroy(&smap_trace_str);

	if (!log_tag)
		log_tag = progname;
	if (log_to_stderr) {
		smap_error_str = smap_ostream_create(flush_stderr_err, NULL);
		smap_debug_str = smap_ostream_create(flush_stderr_dbg, NULL);
		if (smap_trace_option)
			smap_trace_str = smap_ostream_create(flush_stderr_trc,
							     NULL);
	} else {
		openlog(log_tag, LOG_PID, log_facility);
		smap_error_str = smap_ostream_create(flush_syslog,
						     (void*) LOG_ERR);
		smap_debug_str = smap_ostream_create(flush_syslog,
						     (void*) LOG_DEBUG);
		if (smap_trace_option)
			smap_trace_str = smap_ostream_create(flush_syslog,
							     (void*) LOG_INFO);
	}
}

static struct smap_kwtab kw_facility[] = {
	{ "USER",    LOG_USER },
	{ "user",    LOG_USER },
	{ "DAEMON",  LOG_DAEMON },
	{ "daemon",  LOG_DAEMON },
	{ "AUTH",    LOG_AUTH },
	{ "auth",    LOG_AUTH },
	{ "AUTHPRIV",LOG_AUTHPRIV },
	{ "authpriv",LOG_AUTHPRIV },
	{ "MAIL",    LOG_MAIL },
	{ "mail",    LOG_MAIL },
	{ "CRON",    LOG_CRON },
	{ "cron",    LOG_CRON },
	{ "LOCAL0",  LOG_LOCAL0 },
	{ "local0",  LOG_LOCAL0 },
	{ "LOCAL1",  LOG_LOCAL1 },
	{ "local1",  LOG_LOCAL1 },
	{ "LOCAL2",  LOG_LOCAL2 },
	{ "local2",  LOG_LOCAL2 },
	{ "LOCAL3",  LOG_LOCAL3 },
	{ "local3",  LOG_LOCAL3 },
	{ "LOCAL4",  LOG_LOCAL4 },
	{ "local4",  LOG_LOCAL4 },
	{ "LOCAL5",  LOG_LOCAL5 },
	{ "local5",  LOG_LOCAL5 },
	{ "LOCAL6",  LOG_LOCAL6 },
	{ "local6",  LOG_LOCAL6 },
	{ "LOCAL7",  LOG_LOCAL7 },
	{ "local7",  LOG_LOCAL7 },
	{ NULL }
};

int
syslog_fname_to_n (const char *str, int *pint)
{
	return smap_kwtab_nametotok(kw_facility, str, pint);
}


int debug_level[MAX_DEBUG_COUNT];

void
debug_printf(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	smap_ostream_vprintf(smap_debug_str, fmt, ap);
	smap_ostream_write(smap_debug_str, "\n", 1);
	va_end(ap);
}

int
debug_set(const char *spec)
{
	char *p;
	unsigned n, lev;

	n = strtoul(spec, &p, 0);
	if (!*p) {
		int i;

		for (i = 0; i < sizeof(debug_level)/sizeof(debug_level[0]);
		     i++)
			if (debug_level[i] == 0)
				debug_level[i] = n;
		return;
	}

	if (*p != '.')
		return -1;
	
	lev = strtoul(p + 1, &p, 0);
	if (*p)
		return -1;
	debug_level[n] = lev;
	return 0;
}
	
	
