/* This file is part of smap.
   Copyright (C) 2006, 2007, 2010 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "ostr.h"

struct smap_output_stream {
	char *buf;
	size_t size;
	size_t level;
	void (*flush) (void *data, char *buf);
	void *data;
};

static int
_chkbuf(struct smap_output_stream *str, size_t size)
{
	if (str->level + size + 1 > str->size) {
		size_t newsize = str->level + size + 1;
		char *newbuf = realloc(str->buf, newsize);
		if (!newbuf)
			return -1;
		str->size = newsize;
		str->buf = newbuf;
	}
	return 0;
}

static void
_flushbuf(struct smap_output_stream *str)
{
	char *ptr = str->buf;
	size_t size = str->level;
	
	while (size) {
		size_t len;
		char *p = memchr(ptr, '\n', size);
		if (!p)
			break;
		*p++ = 0;
		str->flush(str->data, ptr);
		len = p - ptr;
		ptr += len;
		size -= len;
	}
	if (ptr > str->buf && size)
		memmove(str->buf, ptr, size);
	str->level = size;
}

int
smap_ostream_write(smap_ostream_t str, char *buf, size_t len)
{
	if (!str) {
		errno = EINVAL;
		return -1;
	}
	if (_chkbuf(str, len))
		return -1;
	memcpy(str->buf + str->level, buf, len);
	str->level += len;
	_flushbuf(str);
	return 0;
}

int
smap_ostream_vprintf(smap_ostream_t str, const char *fmt, va_list ap)
{
	int n;
	
	if (!str) {
		errno = EINVAL;
		return -1;
	}
	if (!str->buf) {
		if (str->size == 0)
			str->size = 512; /* Initial allocation */

		str->buf = calloc(1, str->size);
		if (!str->buf)
			return -1;
		str->level = 0;
	}

	for (;;) {
		size_t buflen = str->size - str->level;
		n = vsnprintf(str->buf + str->level, buflen, fmt, ap);
		
		if (n < 0 || n >= buflen
		    || !memchr (str->buf + str->level, '\0', n + 1))	{
			char *newbuf;
			size_t newlen = str->size * 2;
			if (newlen < str->size) {
				errno = ENOMEM;
				return -1;
			}
			newbuf = realloc(str->buf, newlen);
			if (!newbuf)
				return -1;
			str->size = newlen;
			str->buf = newbuf;
		} else {
			str->level += n;
			break;
		}
	}
	_flushbuf(str);
	return n;
}

int
smap_ostream_printf(smap_ostream_t str, const char *fmt, ...)
{
	va_list ap;
	int n;
	
	if (!str) {
		errno = EINVAL;
		return -1;
	}
	va_start(ap, fmt);
	n = smap_ostream_vprintf(str, fmt, ap);
	va_end(ap);
	return n;
}

int
smap_ostream_flush(smap_ostream_t str)
{
	if (!str) {
		errno = EINVAL;
		return -1;
	}
	_flushbuf(str);
	if (str->level) {
		str->buf[str->level] = 0;
		str->flush(str->data, str->buf);
		str->level = 0;
	}
	return 0;
}
	
smap_ostream_t
smap_ostream_create(void (*flush)(void *data, char *buf), void *data)
{
	struct smap_output_stream *str = calloc(1, sizeof(str[0]));
	if (!str)
		return NULL;
	str->flush = flush;
	str->data = data;
	str->size = 512;
	str->buf = malloc(str->size);
	if (!str->buf) {
		free(str);
		return NULL;
	}
	return str;
}

void
smap_ostream_free(smap_ostream_t str)
{
	if (!str)
		return;
	free(str->buf);
	free(str);
}

void
smap_ostream_destroy(smap_ostream_t *str)
{
	if (!str)
		return;
	if (*str) {
		smap_ostream_free(*str);
		*str = NULL;
	}
}
